@@ -54,17 +54,25 @@ for symbol in exported_pycares_symbols:
 
 exported_pycares_symbols_map = {
     # Query types
-    "QUERY_TYPE_A" : "T_A",
-    "QUERY_TYPE_AAAA" : "T_AAAA",
-    "QUERY_TYPE_ANY" : "T_ANY",
+    "QUERY_TYPE_A"     : "T_A",
+    "QUERY_TYPE_AAAA"  : "T_AAAA",
+    "QUERY_TYPE_ANY"   : "T_ANY",
+    "QUERY_TYPE_CAA"   : "T_CAA",
     "QUERY_TYPE_CNAME" : "T_CNAME",
-    "QUERY_TYPE_MX" : "T_MX",
+    "QUERY_TYPE_MX"    : "T_MX",
     "QUERY_TYPE_NAPTR" : "T_NAPTR",
-    "QUERY_TYPE_NS" : "T_NS",
-    "QUERY_TYPE_PTR" : "T_PTR",
-    "QUERY_TYPE_SOA" : "T_SOA",
-    "QUERY_TYPE_SRV" : "T_SRV",
-    "QUERY_TYPE_TXT" : "T_TXT",
+    "QUERY_TYPE_NS"    : "T_NS",
+    "QUERY_TYPE_PTR"   : "T_PTR",
+    "QUERY_TYPE_SOA"   : "T_SOA",
+    "QUERY_TYPE_SRV"   : "T_SRV",
+    "QUERY_TYPE_TXT"   : "T_TXT",
+
+    # Query classes
+    "QUERY_CLASS_IN"   : "C_IN",
+    "QUERY_CLASS_CHAOS": "C_CHAOS",
+    "QUERY_CLASS_HS"   : "C_HS",
+    "QUERY_CLASS_NONE" :"C_NONE",
+    "QUERY_CLASS_ANY"  : "C_ANY",
 }
 
 for k, v in exported_pycares_symbols_map.items():
@@ -85,12 +93,12 @@ class AresError(Exception):
 
 _global_set = set()
 
-@_ffi.callback("void (void *data, ares_socket_t socket_fd, int readable, int writable )")
+@_ffi.def_extern()
 def _sock_state_cb(data, socket_fd, readable, writable):
     sock_state_cb = _ffi.from_handle(data)
     sock_state_cb(socket_fd, readable, writable)
 
-@_ffi.callback("void (void *arg, int status, int timeouts, struct hostent *hostent)")
+@_ffi.def_extern()
 def _host_cb(arg, status, timeouts, hostent):
     callback = _ffi.from_handle(arg)
     _global_set.discard(arg)
@@ -103,7 +111,7 @@ def _host_cb(arg, status, timeouts, hostent):
 
     callback(result, status)
 
-@_ffi.callback("void (void *arg, int status, int timeouts, char *node, char *service)")
+@_ffi.def_extern()
 def _nameinfo_cb(arg, status, timeouts, node, service):
     callback = _ffi.from_handle(arg)
     _global_set.discard(arg)
@@ -116,7 +124,7 @@ def _nameinfo_cb(arg, status, timeouts, node, service):
 
     callback(result, status)
 
-@_ffi.callback("void (void *arg, int status, int timeouts, unsigned char *abuf, int alen)")
+@_ffi.def_extern()
 def _query_cb(arg, status, timeouts, abuf, alen):
     callback, query_type = _ffi.from_handle(arg)
     _global_set.discard(arg)
@@ -124,7 +132,7 @@ def _query_cb(arg, status, timeouts, abuf, alen):
     if status == _lib.ARES_SUCCESS:
         if query_type == _lib.T_ANY:
             result = []
-            for qtype in (_lib.T_A, _lib.T_AAAA, _lib.T_CNAME, _lib.T_MX, _lib.T_NAPTR, _lib.T_NS, _lib.T_PTR, _lib.T_SOA, _lib.T_SRV, _lib.T_TXT):
+            for qtype in (_lib.T_A, _lib.T_AAAA, _lib.T_CAA, _lib.T_CNAME, _lib.T_MX, _lib.T_NAPTR, _lib.T_NS, _lib.T_PTR, _lib.T_SOA, _lib.T_SRV, _lib.T_TXT):
                 r, status = parse_result(qtype, abuf, alen)
                 if status not in (None, _lib.ARES_ENODATA, _lib.ARES_EBADRESP):
                     result = None
@@ -143,6 +151,19 @@ def _query_cb(arg, status, timeouts, abuf, alen):
 
     callback(result, status)
 
+@_ffi.def_extern()
+def _addrinfo_cb(arg, status, timeouts, res):
+    callback = _ffi.from_handle(arg)
+    _global_set.discard(arg)
+
+    if status != _lib.ARES_SUCCESS:
+        result = None
+    else:
+        result = ares_addrinfo_result(res)
+        status = None
+
+    callback(result, status)
+
 def parse_result(query_type, abuf, alen):
     if query_type == _lib.T_A:
         addrttls = _ffi.new("struct ares_addrttl[]", PYCARES_ADDRTTL_SIZE)
@@ -164,6 +185,20 @@ def parse_result(query_type, abuf, alen):
         else:
             result = [ares_query_aaaa_result(addrttls[i]) for i in range(naddrttls[0])]
             status = None
+    elif query_type == _lib.T_CAA:
+        caa_reply = _ffi.new("struct ares_caa_reply **")
+        parse_status = _lib.ares_parse_caa_reply(abuf, alen, caa_reply)
+        if parse_status != _lib.ARES_SUCCESS:
+            result = None
+            status = parse_status
+        else:
+            result = []
+            caa_reply_ptr = caa_reply[0]
+            while caa_reply_ptr != _ffi.NULL:
+                result.append(ares_query_caa_result(caa_reply_ptr))
+                caa_reply_ptr = caa_reply_ptr.next
+            _lib.ares_free_data(caa_reply[0])
+            status = None
     elif query_type == _lib.T_CNAME:
         host = _ffi.new("struct hostent **")
         parse_status = _lib.ares_parse_a_reply(abuf, alen, host, _ffi.NULL, _ffi.NULL)
@@ -176,7 +211,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_MX:
         mx_reply = _ffi.new("struct ares_mx_reply **")
-        parse_status = _lib.ares_parse_mx_reply(abuf, alen, mx_reply);
+        parse_status = _lib.ares_parse_mx_reply(abuf, alen, mx_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -190,7 +225,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_NAPTR:
         naptr_reply = _ffi.new("struct ares_naptr_reply **")
-        parse_status = _lib.ares_parse_naptr_reply(abuf, alen, naptr_reply);
+        parse_status = _lib.ares_parse_naptr_reply(abuf, alen, naptr_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -204,7 +239,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_NS:
         hostent = _ffi.new("struct hostent **")
-        parse_status = _lib.ares_parse_ns_reply(abuf, alen, hostent);
+        parse_status = _lib.ares_parse_ns_reply(abuf, alen, hostent)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -219,8 +254,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_PTR:
         hostent = _ffi.new("struct hostent **")
-        hostttl = _ffi.new("int*", PYCARES_ADDRTTL_SIZE)
-        parse_status = _lib.ares_parse_ptr_reply(abuf, alen, _ffi.NULL, 0, socket.AF_UNSPEC, hostent, hostttl);
+        parse_status = _lib.ares_parse_ptr_reply(abuf, alen, _ffi.NULL, 0, socket.AF_UNSPEC, hostent)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -231,12 +265,12 @@ def parse_result(query_type, abuf, alen):
             while host.h_aliases[i] != _ffi.NULL:
                 aliases.append(maybe_str(_ffi.string(host.h_aliases[i])))
                 i += 1
-            result = ares_query_ptr_result(host, hostttl[0], aliases)
+            result = ares_query_ptr_result(host, aliases)
             _lib.ares_free_hostent(host)
             status = None
     elif query_type == _lib.T_SOA:
         soa_reply = _ffi.new("struct ares_soa_reply **")
-        parse_status = _lib.ares_parse_soa_reply(abuf, alen, soa_reply);
+        parse_status = _lib.ares_parse_soa_reply(abuf, alen, soa_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -246,7 +280,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_SRV:
         srv_reply = _ffi.new("struct ares_srv_reply **")
-        parse_status = _lib.ares_parse_srv_reply(abuf, alen, srv_reply);
+        parse_status = _lib.ares_parse_srv_reply(abuf, alen, srv_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -260,7 +294,7 @@ def parse_result(query_type, abuf, alen):
             status = None
     elif query_type == _lib.T_TXT:
         txt_reply = _ffi.new("struct ares_txt_ext **")
-        parse_status = _lib.ares_parse_txt_reply_ext(abuf, alen, txt_reply);
+        parse_status = _lib.ares_parse_txt_reply_ext(abuf, alen, txt_reply)
         if parse_status != _lib.ARES_SUCCESS:
             result = None
             status = parse_status
@@ -271,14 +305,14 @@ def parse_result(query_type, abuf, alen):
             while True:
                 if txt_reply_ptr == _ffi.NULL:
                     if tmp_obj is not None:
-                        result.append(tmp_obj)
+                        result.append(ares_query_txt_result(tmp_obj))
                     break
                 if txt_reply_ptr.record_start == 1:
                     if tmp_obj is not None:
-                        result.append(tmp_obj)
-                    tmp_obj = ares_query_txt_result(txt_reply_ptr)
+                        result.append(ares_query_txt_result(tmp_obj))
+                    tmp_obj = ares_query_txt_result_chunk(txt_reply_ptr)
                 else:
-                    new_chunk = ares_query_txt_result(txt_reply_ptr)
+                    new_chunk = ares_query_txt_result_chunk(txt_reply_ptr)
                     tmp_obj.text += new_chunk.text
                 txt_reply_ptr = txt_reply_ptr.next
             _lib.ares_free_data(txt_reply[0])
@@ -290,7 +324,8 @@ def parse_result(query_type, abuf, alen):
 
 
 class Channel:
-    __qtypes__ = (_lib.T_A, _lib.T_AAAA, _lib.T_ANY, _lib.T_CNAME, _lib.T_MX, _lib.T_NAPTR, _lib.T_NS, _lib.T_PTR, _lib.T_SOA, _lib.T_SRV, _lib.T_TXT)
+    __qtypes__ = (_lib.T_A, _lib.T_AAAA, _lib.T_ANY, _lib.T_CAA, _lib.T_CNAME, _lib.T_MX, _lib.T_NAPTR, _lib.T_NS, _lib.T_PTR, _lib.T_SOA, _lib.T_SRV, _lib.T_TXT)
+    __qclasses__ = (_lib.C_IN, _lib.C_CHAOS, _lib.C_HS, _lib.C_NONE, _lib.C_ANY)
 
     def __init__(self,
                  flags = None,
@@ -355,12 +390,12 @@ class Channel:
             # This must be kept alive while the channel is alive.
             self._sock_state_cb_handle = userdata
 
-            options.sock_state_cb = _sock_state_cb
+            options.sock_state_cb = _lib._sock_state_cb
             options.sock_state_cb_data = userdata
             optmask = optmask |  _lib.ARES_OPT_SOCK_STATE_CB
 
         if lookups:
-            options.lookups = lookups
+            options.lookups = _ffi.new('char[]', ascii_bytes(lookups))
             optmask = optmask |  _lib.ARES_OPT_LOOKUPS
 
         if domains:
@@ -491,7 +526,7 @@ class Channel:
 
         userdata = _ffi.new_handle(callback)
         _global_set.add(userdata)
-        _lib.ares_gethostbyaddr(self._channel[0], address, _ffi.sizeof(address[0]), family, _host_cb, userdata)
+        _lib.ares_gethostbyaddr(self._channel[0], address, _ffi.sizeof(address[0]), family, _lib._host_cb, userdata)
 
     def gethostbyname(self, name, family, callback):
         if not callable(callback):
@@ -499,24 +534,51 @@ class Channel:
 
         userdata = _ffi.new_handle(callback)
         _global_set.add(userdata)
-        _lib.ares_gethostbyname(self._channel[0], parse_name(name), family, _host_cb, userdata)
+        _lib.ares_gethostbyname(self._channel[0], parse_name(name), family, _lib._host_cb, userdata)
 
-    def query(self, name, query_type, callback):
-        self._do_query(_lib.ares_query, name, query_type, callback)
+    def getaddrinfo(self, host, port, callback, family=0, type=0, proto=0, flags=0):
+        if not callable(callback):
+            raise TypeError("a callable is required")
 
-    def search(self, name, query_type, callback):
-        self._do_query(_lib.ares_search, name, query_type, callback)
+        if port is None:
+            service = _ffi.NULL
+        elif isinstance(port, int):
+            service = str(port).encode('ascii')
+        else:
+            service = ascii_bytes(port)
 
-    def _do_query(self, func, name, query_type, callback):
+        userdata = _ffi.new_handle(callback)
+        _global_set.add(userdata)
+
+        hints = _ffi.new('struct ares_addrinfo_hints*')
+        hints.ai_flags = flags
+        hints.ai_family = family
+        hints.ai_socktype = type
+        hints.ai_protocol = proto
+        _lib.ares_getaddrinfo(self._channel[0], parse_name(host), service, hints, _lib._addrinfo_cb, userdata)
+
+    def query(self, name, query_type, callback, query_class=None):
+        self._do_query(_lib.ares_query, name, query_type, callback, query_class=query_class)
+
+    def search(self, name, query_type, callback, query_class=None):
+        self._do_query(_lib.ares_search, name, query_type, callback, query_class=query_class)
+
+    def _do_query(self, func, name, query_type, callback, query_class=None):
         if not callable(callback):
             raise TypeError('a callable is required')
 
         if query_type not in self.__qtypes__:
             raise ValueError('invalid query type specified')
 
+        if query_class is None:
+            query_class = _lib.C_IN
+
+        if query_class not in self.__qclasses__:
+            raise ValueError('invalid query class specified')
+
         userdata = _ffi.new_handle((callback, query_type))
         _global_set.add(userdata)
-        func(self._channel[0], parse_name(name), _lib.C_IN, query_type, _query_cb, userdata)
+        func(self._channel[0], parse_name(name), query_class, query_type, _lib._query_cb, userdata)
 
     def set_local_ip(self, ip):
         addr4 = _ffi.new("struct in_addr*")
@@ -528,32 +590,34 @@ class Channel:
         else:
             raise ValueError("invalid IP address")
 
-    def getnameinfo(self, ip_port, flags, callback):
+    def getnameinfo(self, address, flags, callback):
         if not callable(callback):
             raise TypeError("a callable is required")
 
-        ip, port = ip_port
-
-        if port < 0 or port > 65535:
-            raise ValueError("port must be between 0 and 65535")
-
-        sa4 = _ffi.new("struct sockaddr_in*")
-        sa6 = _ffi.new("struct sockaddr_in6*")
-
-        if _lib.ares_inet_pton(socket.AF_INET, ascii_bytes(ip), _ffi.addressof(sa4.sin_addr)) == 1:
+        if len(address) == 2:
+            (ip, port) = address
+            sa4 = _ffi.new("struct sockaddr_in*")
+            if _lib.ares_inet_pton(socket.AF_INET, ascii_bytes(ip), _ffi.addressof(sa4.sin_addr)) != 1:
+                raise ValueError("Invalid IPv4 address %r" % ip)
             sa4.sin_family = socket.AF_INET
             sa4.sin_port = socket.htons(port)
             sa = sa4
-        elif _lib.ares_inet_pton(socket.AF_INET6, ascii_bytes(ip), _ffi.addressof(sa6.sin6_addr)) == 1:
+        elif len(address) == 4:
+            (ip, port, flowinfo, scope_id) = address
+            sa6 = _ffi.new("struct sockaddr_in6*")
+            if _lib.ares_inet_pton(socket.AF_INET6, ascii_bytes(ip), _ffi.addressof(sa6.sin6_addr)) != 1:
+                raise ValueError("Invalid IPv6 address %r" % ip)
             sa6.sin6_family = socket.AF_INET6
             sa6.sin6_port = socket.htons(port)
+            sa6.sin6_flowinfo = socket.htonl(flowinfo) # I'm unsure about byteorder here.
+            sa6.sin6_scope_id = scope_id # Yes, without htonl.
             sa = sa6
         else:
-            raise ValueError("invalid IP address")
+            raise ValueError("Invalid address argument")
 
         userdata = _ffi.new_handle(callback)
         _global_set.add(userdata)
-        _lib.ares_getnameinfo(self._channel[0], _ffi.cast("struct sockaddr*", sa), _ffi.sizeof(sa[0]), flags, _nameinfo_cb, userdata)
+        _lib.ares_getnameinfo(self._channel[0], _ffi.cast("struct sockaddr*", sa), _ffi.sizeof(sa[0]), flags, _lib._nameinfo_cb, userdata)
 
     def set_local_dev(self, dev):
         _lib.ares_set_local_dev(self._channel[0], dev)
@@ -592,6 +656,17 @@ class ares_query_aaaa_result(AresResult):
         self.ttl = ares_addrttl.ttl
 
 
+class  ares_query_caa_result(AresResult):
+    __slots__ = ('critical', 'property', 'value', 'ttl')
+    type = 'CAA'
+
+    def __init__(self, caa):
+        self.critical = caa.critical
+        self.property = maybe_str(_ffi.string(caa.property, caa.plength))
+        self.value = maybe_str(_ffi.string(caa.value, caa.length))
+        self.ttl = -1
+
+
 class ares_query_cname_result(AresResult):
     __slots__ = ('cname', 'ttl')
     type = 'CNAME'
@@ -608,7 +683,7 @@ class ares_query_mx_result(AresResult):
     def __init__(self, mx):
         self.host = maybe_str(_ffi.string(mx.host))
         self.priority = mx.priority
-        self.ttl = mx.ttl
+        self.ttl = -1
 
 
 class ares_query_naptr_result(AresResult):
@@ -622,7 +697,7 @@ class ares_query_naptr_result(AresResult):
         self.service = maybe_str(_ffi.string(naptr.service))
         self.regex = maybe_str(_ffi.string(naptr.regexp))
         self.replacement = maybe_str(_ffi.string(naptr.replacement))
-        self.ttl = naptr.ttl
+        self.ttl = -1
 
 
 class ares_query_ns_result(AresResult):
@@ -638,10 +713,10 @@ class ares_query_ptr_result(AresResult):
     __slots__ = ('name', 'ttl', 'aliases')
     type = 'PTR'
 
-    def __init__(self, hostent, ttl, aliases):
+    def __init__(self, hostent, aliases):
         self.name = maybe_str(_ffi.string(hostent.h_name))
-        self.ttl = ttl
         self.aliases = aliases
+        self.ttl = -1
 
 
 class ares_query_soa_result(AresResult):
@@ -656,7 +731,7 @@ class ares_query_soa_result(AresResult):
         self.retry = soa.retry
         self.expires = soa.expire
         self.minttl = soa.minttl
-        self.ttl = soa.ttl
+        self.ttl = -1
 
 
 class  ares_query_srv_result(AresResult):
@@ -668,16 +743,25 @@ class  ares_query_srv_result(AresResult):
         self.port = srv.port
         self.priority = srv.priority
         self.weight = srv.weight
-        self.ttl = srv.ttl
+        self.ttl = -1
 
 
 class ares_query_txt_result(AresResult):
     __slots__ = ('text', 'ttl')
     type = 'TXT'
 
+    def __init__(self, txt_chunk):
+        self.text = maybe_str(txt_chunk.text)
+        self.ttl = -1
+
+
+class ares_query_txt_result_chunk(AresResult):
+    __slots__ = ('text', 'ttl')
+    type = 'TXT'
+
     def __init__(self, txt):
-        self.text = maybe_str(_ffi.string(txt.txt))
-        self.ttl = txt.ttl
+        self.text = _ffi.string(txt.txt)
+        self.ttl = -1
 
 
 # Other result types
@@ -711,7 +795,61 @@ class ares_nameinfo_result(AresResult):
         self.service = maybe_str(_ffi.string(service)) if service != _ffi.NULL else None
 
 
+class ares_addrinfo_node_result(AresResult):
+    __slots__ = ('ttl', 'flags', 'family', 'socktype', 'protocol', 'addr')
+
+    def __init__(self, ares_node):
+        self.ttl = ares_node.ai_ttl
+        self.flags = ares_node.ai_flags
+        self.socktype = ares_node.ai_socktype
+        self.protocol = ares_node.ai_protocol
+
+        addr = ares_node.ai_addr
+        assert addr.sa_family == ares_node.ai_family
+        ip = _ffi.new("char []", _lib.INET6_ADDRSTRLEN)
+        if addr.sa_family == socket.AF_INET:
+            self.family = socket.AF_INET
+            s = _ffi.cast("struct sockaddr_in*", addr)
+            if _ffi.NULL != _lib.ares_inet_ntop(s.sin_family, _ffi.addressof(s.sin_addr), ip, _lib.INET6_ADDRSTRLEN):
+                # (address, port) 2-tuple for AF_INET
+                self.addr = (_ffi.string(ip, _lib.INET6_ADDRSTRLEN), socket.ntohs(s.sin_port))
+        elif addr.sa_family == socket.AF_INET6:
+            self.family = socket.AF_INET6
+            s = _ffi.cast("struct sockaddr_in6*", addr)
+            if _ffi.NULL != _lib.ares_inet_ntop(s.sin6_family, _ffi.addressof(s.sin6_addr), ip, _lib.INET6_ADDRSTRLEN):
+                # (address, port, flow info, scope id) 4-tuple for AF_INET6
+                self.addr = (_ffi.string(ip, _lib.INET6_ADDRSTRLEN), socket.ntohs(s.sin6_port), s.sin6_flowinfo, s.sin6_scope_id)
+        else:
+            raise ValueError("invalid sockaddr family")
+
+
+class ares_addrinfo_cname_result(AresResult):
+    __slots__ = ('ttl', 'alias', 'name')
+
+    def __init__(self, ares_cname):
+        self.ttl = ares_cname.ttl
+        self.alias = maybe_str(_ffi.string(ares_cname.alias))
+        self.name = maybe_str(_ffi.string(ares_cname.name))
+
+
+class ares_addrinfo_result(AresResult):
+    __slots__ = ('cnames', 'nodes')
+
+    def __init__(self, ares_addrinfo):
+        self.cnames = []
+        self.nodes = []
+        cname_ptr = ares_addrinfo.cnames
+        while cname_ptr != _ffi.NULL:
+            self.cnames.append(ares_addrinfo_cname_result(cname_ptr))
+            cname_ptr = cname_ptr.next
+        node_ptr = ares_addrinfo.nodes
+        while node_ptr != _ffi.NULL:
+            self.nodes.append(ares_addrinfo_node_result(node_ptr))
+            node_ptr = node_ptr.ai_next
+        _lib.ares_freeaddrinfo(ares_addrinfo)
+
+
+
 __all__ = exported_pycares_symbols + list(exported_pycares_symbols_map.keys()) + ['AresError', 'Channel', 'errno', '__version__']
 
 del exported_pycares_symbols, exported_pycares_symbols_map
-
